from .base import PluginMapping, PluginProtocol, get_plugin_for_value
from .sql_alchemy import SQLAlchemyPlugin

__all__ = [
    "PluginMapping",
    "PluginProtocol",
    "SQLAlchemyPlugin",
    "get_plugin_for_value",
]



import os
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    Iterator,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from openapi_schema_pydantic import Header
from pydantic import BaseModel, FilePath, create_model, validator
from pydantic.typing import AnyCallable
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.requests import HTTPConnection
from starlette.responses import Response as StarletteResponse
from starlette.types import ASGIApp, Receive, Scope, Send
from typing_extensions import AsyncIterator, Literal, Protocol, Type, runtime_checkable

from starlite.exceptions import HTTPException
from starlite.response import Response

try:
    # python 3.9 changed these variable
    from typing import _UnionGenericAlias as GenericAlias  # type: ignore
except ImportError:  # pragma: no cover
    from typing import _GenericAlias as GenericAlias  # type: ignore

if TYPE_CHECKING:  # pragma: no cover
    from starlite.controller import Controller
    from starlite.handlers import BaseRouteHandler
    from starlite.request import Request
    from starlite.routing import Router
else:
    Request = Any
    BaseRouteHandler = Any
    Controller = Any
    Router = Any

ExceptionHandler = Callable[
    [Request, Union[Exception, HTTPException, StarletteHTTPException]], Union[Response, StarletteResponse]
]
Guard = Union[
    Callable[[HTTPConnection, BaseRouteHandler], Awaitable[None]], Callable[[HTTPConnection, BaseRouteHandler], None]
]
Method = Union[Literal["GET"], Literal["POST"], Literal["DELETE"], Literal["PATCH"], Literal["PUT"], Literal["HEAD"]]
ControllerRouterHandler = Union[Type[Controller], BaseRouteHandler, Router, AnyCallable]


@runtime_checkable
class MiddlewareProtocol(Protocol):
    def __init__(self, app: ASGIApp):  # pragma: no cover
        ...

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:  # pragma: no cover
        ...


T = TypeVar("T", bound=Type[BaseModel])


class Partial(Generic[T]):
    _models: Dict[T, Any] = {}

    def __class_getitem__(cls, item: T) -> T:
        """
        Modifies a given T subclass of BaseModel to be all optional
        """
        if not cls._models.get(item):
            field_definitions: Dict[str, Tuple[Any, None]] = {}
            for field_name, field_type in item.__annotations__.items():
                # we modify the field annotations to make it optional
                if not isinstance(field_type, GenericAlias) or type(None) not in field_type.__args__:
                    field_definitions[field_name] = (Optional[field_type], None)
                else:
                    field_definitions[field_name] = (field_type, None)
                cls._models[item] = create_model("Partial" + item.__name__, **field_definitions)  # type: ignore
        return cast(T, cls._models.get(item))


class File(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    path: FilePath
    filename: str
    stat_result: Optional[os.stat_result] = None

    @validator("stat_result", always=True)
    def validate_status_code(  # pylint: disable=no-self-argument,no-self-use
        cls, value: Optional[os.stat_result], values: Dict[str, Any]
    ) -> os.stat_result:
        """Set the stat_result value for the given filepath"""
        return value or os.stat(cast(str, values.get("path")))


class Redirect(BaseModel):
    path: str


class Stream(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    iterator: Union[Iterator[Any], AsyncIterator[Any]]


class ResponseHeader(Header):  # type: ignore
    value: Any = ...



# flake8: noqa
from .model import convert_dataclass_to_model, create_parsed_model_field
from .sequence import find_index, unique
from .signature import SignatureModel, create_function_signature_model
from .url import join_paths, normalize_path

__all__ = [
    "SignatureModel",
    "convert_dataclass_to_model",
    "create_function_signature_model",
    "create_parsed_model_field",
    "find_index",
    "join_paths",
    "normalize_path",
    "unique",
]



from dataclasses import is_dataclass
from typing import Any, Dict, ForwardRef, List, Optional, Tuple, Union, cast

from pydantic import BaseModel, create_model
from pydantic.fields import ModelField, Undefined
from typing_extensions import Type

from starlite.exceptions import ImproperlyConfiguredException
from starlite.plugins import PluginProtocol, get_plugin_for_value
from starlite.utils import convert_dataclass_to_model


class DTOFactory:
    def __init__(self, plugins: Optional[List[PluginProtocol]] = None):
        self.plugins = plugins or []

    def __call__(
        self,
        name: str,
        source: Any,
        exclude: Optional[List[str]] = None,
        field_mapping: Optional[Dict[str, Union[str, Tuple[str, Any]]]] = None,
    ) -> Type[BaseModel]:
        """
        Given a supported model class - either pydantic, dataclass or a class supported via plugins,
        create a DTO pydantic model class.

        An instance of the factory must first be created, passing any plugins to it.
        It can then be used to create a DTO by calling the instance like a function. Additionally, it can exclude (drop)
        attributes specifies in the 'exclude' list and remap field names and/or field types.

        For example, given a pydantic model

            class MyClass(BaseModel):
                first: int
                second: int

            MyClassDTO = DTOFactory()(MyClass, exclude=["first"], field_mapping={"second": ("third", float)})

        `MyClassDTO` is now equal to this:
            class MyClassDTO(BaseModel):
                third: float

        It can be used as a regular pydantic model:

            @post(path="/my-path)
            def create_obj(data: MyClassDTO) -> MyClass:
                ...

        This will affect parsing, validation and how OpenAPI schema is generated exactly like when using a pydantic model.

        Note: Although the value generated is a pydantic factory, because it is being generated programmaticaly,
        it's currently not possible to extend editor auto-complete for the DTO properties - it will be typed as a
        Pydantic BaseModel, but no attributes will be inferred in the editor.
        """
        fields: Dict[str, ModelField]
        exclude = exclude or []
        field_mapping = field_mapping or {}
        field_definitions: Dict[str, Tuple[Any, Any]] = {}
        if issubclass(source, BaseModel):
            source.update_forward_refs()
            fields = source.__fields__
        elif is_dataclass(source):
            fields = convert_dataclass_to_model(source).__fields__
        else:
            plugin = get_plugin_for_value(value=source, plugins=self.plugins)
            if not plugin:
                raise ImproperlyConfiguredException(
                    f"No supported plugin found for value {source} - cannot create value"
                )
            model = plugin.to_pydantic_model_class(model_class=source)
            fields = model.__fields__
        for field_name, model_field in fields.items():
            if field_name not in exclude:
                outer_type = model_field.outer_type_
                field_type = outer_type if not isinstance(outer_type, ForwardRef) else model_field.type_
                if field_name in field_mapping:
                    mapping = field_mapping[field_name]
                    if isinstance(mapping, tuple):
                        field_name, field_type = mapping
                    else:
                        field_name = mapping
                if model_field.field_info.default is not Undefined:
                    field_definitions[field_name] = (field_type, model_field.default)
                elif not model_field.allow_none:
                    field_definitions[field_name] = (field_type, ...)
                else:
                    field_definitions[field_name] = (field_type, None)
        return cast(Type[BaseModel], create_model(name, **field_definitions))  # type: ignore



from openapi_schema_pydantic import OpenAPI
from orjson import OPT_INDENT_2, dumps

from starlite.controller import Controller
from starlite.enums import MediaType, OpenAPIMediaType
from starlite.exceptions import ImproperlyConfiguredException
from starlite.handlers import get
from starlite.request import Request


class OpenAPIController(Controller):
    path = "/schema"
    styles = """
        body { margin: 0; padding: 0 }
    """
    redoc_version = "next"

    @staticmethod
    def schema_from_request(request: Request) -> OpenAPI:
        """Returns the openapi schema"""
        if not request.app.openapi_schema:  # pragma: no cover
            raise ImproperlyConfiguredException("Starlite has not been instantiated with OpenAPIConfig")
        return request.app.openapi_schema

    @get(path="/openapi.yaml", media_type=OpenAPIMediaType.OPENAPI_YAML, include_in_schema=False)
    def retrieve_schema_yaml(self, request: Request) -> OpenAPI:
        """Returns the openapi schema"""
        return self.schema_from_request(request)

    @get(path="/openapi.json", media_type=OpenAPIMediaType.OPENAPI_JSON, include_in_schema=False)
    def retrieve_schema_json(self, request: Request) -> OpenAPI:
        """Returns the openapi schema"""
        return self.schema_from_request(request)

    @get(path="/redoc", media_type=MediaType.HTML, include_in_schema=False)
    def redoc(self, request: Request) -> str:  # pragma: no cover
        """Endpoint that serves Redoc"""
        dumped_schema = dumps(
            request.app.openapi_schema.json(by_alias=True, exclude_none=True), option=OPT_INDENT_2
        ).decode("utf-8")
        head = f"""
          <head>
            <title>{request.app.openapi_schema.info.title}</title>
            <meta charset="utf-8"/>
            <meta name="viewport" content="width=device-width, initial-scale=1">
            <link href="https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700" rel="stylesheet">
            <script src="https://cdn.jsdelivr.net/npm/redoc@{self.redoc_version}/bundles/redoc.standalone.js"></script>
            <style>
                {self.styles}
            </style>
          </head>
        """
        body = f"""
          <body>
            <div id='redoc-container'/>
            <script type="text/javascript">
                Redoc.init(
                    JSON.parse({dumped_schema}),
                    undefined,
                    document.getElementById('redoc-container')
                )
            </script>
          </body>
        """
        return f"""
        <!DOCTYPE html>
            <html>
                {head}
                {body}
            </html>
        """



from typing import Callable, Iterable, List, TypeVar

T = TypeVar("T")


def find_index(target_list: List[T], predicate: Callable[[T], bool]) -> int:
    """Find element in list given a key and value. List elements can be dicts or classes"""
    for i, element in enumerate(target_list):
        if predicate(element):
            return i
    return -1


def unique(value: Iterable[T]) -> List[T]:
    """Return all unique values in a given sequence or iterator"""
    try:
        return list(set(value))
    except TypeError:
        output: List[T] = []
        for element in value:
            if not any(v == element for v in output):
                output.append(element)
        return output



from datetime import date, datetime, time, timedelta
from decimal import Decimal
from inspect import isclass
from ipaddress import IPv4Network, IPv6Network
from typing import Any, Callable, Dict, List, Tuple, Union
from uuid import UUID

from pydantic import BaseModel, Json, conint, constr, create_model
from typing_extensions import Type

from starlite.exceptions import (
    ImproperlyConfiguredException,
    MissingDependencyException,
)
from starlite.plugins.base import PluginProtocol

try:
    from sqlalchemy import Table, inspect
    from sqlalchemy import types as sqlalchemy_type
    from sqlalchemy.dialects import (
        firebird,
        mssql,
        mysql,
        oracle,
        postgresql,
        sqlite,
        sybase,
    )
    from sqlalchemy.orm import DeclarativeMeta, Mapper
    from sqlalchemy.sql.type_api import TypeEngine
except ImportError as exc:  # pragma: no cover
    raise MissingDependencyException("sqlalchemy is not installed") from exc


class SQLAlchemyPlugin(PluginProtocol[Union[DeclarativeMeta, Table]]):
    def __init__(self) -> None:
        # a map object that maps SQLAlchemy entity qualnames to pydantic BaseModel subclasses
        self.model_namespace_map: Dict[str, Type[BaseModel]] = {}

    @staticmethod
    def is_plugin_supported_type(value: Any) -> bool:
        """
        This plugin supports only SQLAlchemy declarative models
        """
        return isinstance(value, DeclarativeMeta) or isinstance(value.__class__, DeclarativeMeta)

    @staticmethod
    def handle_string_type(column_type: Union[sqlalchemy_type.String, sqlalchemy_type._Binary]) -> Type:
        """
        Handles the SQLAlchemy String types, including Blob and Binaric types
        """
        if column_type.length is not None:
            return constr(max_length=column_type.length)
        return str

    @staticmethod
    def handle_numeric_type(column_type: sqlalchemy_type.Numeric) -> Type:
        """
        Handles the SQLAlchemy non-int Numeric types
        """
        if column_type.asdecimal:
            return Decimal
        return float

    def handle_list_type(self, column_type: sqlalchemy_type.ARRAY) -> Any:
        """
        Handles the SQLAlchemy Array type
        """
        list_type: Any = self.get_pydantic_type(column_type=column_type.item_type)

        dimensions = column_type.dimensions or 1
        while dimensions > 0:
            list_type = List[list_type]
            dimensions -= 1
        return list_type

    def handle_tuple_type(self, column_type: sqlalchemy_type.TupleType) -> Any:
        """
        Handles the SQLAlchemy Tuple type
        """
        types = [self.get_pydantic_type(column_type=t) for t in column_type.types]
        return Tuple[tuple(types)]

    def handle_enum(self, column_type: Union[sqlalchemy_type.Enum, mysql.ENUM, postgresql.ENUM]) -> Any:
        """
        Handles the SQLAlchemy Enum types
        """
        return column_type.enum_class

    @property
    def providers_map(self) -> Dict[Type[TypeEngine], Callable[[Union[TypeEngine, Type[TypeEngine]]], Any]]:
        """
        A map of SQLAlchemy column types to provider functions.

        This method is separated to allow for easy overriding in subclasses.
        """
        return {
            sqlalchemy_type.ARRAY: self.handle_list_type,
            sqlalchemy_type.BIGINT: lambda x: int,
            sqlalchemy_type.BINARY: self.handle_string_type,
            sqlalchemy_type.BLOB: self.handle_string_type,
            sqlalchemy_type.BOOLEAN: lambda x: bool,
            sqlalchemy_type.BigInteger: lambda x: int,
            sqlalchemy_type.Boolean: lambda x: bool,
            sqlalchemy_type.CHAR: self.handle_string_type,
            sqlalchemy_type.CLOB: self.handle_string_type,
            sqlalchemy_type.DATE: lambda x: date,
            sqlalchemy_type.DATETIME: lambda x: datetime,
            sqlalchemy_type.DECIMAL: self.handle_numeric_type,
            sqlalchemy_type.Date: lambda x: date,
            sqlalchemy_type.DateTime: lambda x: datetime,
            sqlalchemy_type.Enum: self.handle_enum,
            sqlalchemy_type.FLOAT: self.handle_numeric_type,
            sqlalchemy_type.Float: self.handle_numeric_type,
            sqlalchemy_type.INT: lambda x: int,
            sqlalchemy_type.INTEGER: lambda x: int,
            sqlalchemy_type.Integer: lambda x: int,
            sqlalchemy_type.Interval: lambda x: timedelta,
            sqlalchemy_type.JSON: lambda x: dict,
            sqlalchemy_type.LargeBinary: self.handle_string_type,
            sqlalchemy_type.NCHAR: self.handle_string_type,
            sqlalchemy_type.NUMERIC: self.handle_numeric_type,
            sqlalchemy_type.NVARCHAR: self.handle_string_type,
            sqlalchemy_type.Numeric: self.handle_numeric_type,
            sqlalchemy_type.REAL: self.handle_numeric_type,
            sqlalchemy_type.SMALLINT: lambda x: int,
            sqlalchemy_type.SmallInteger: lambda x: int,
            sqlalchemy_type.String: self.handle_string_type,
            sqlalchemy_type.TEXT: self.handle_string_type,
            sqlalchemy_type.TIME: lambda x: time,
            sqlalchemy_type.TIMESTAMP: lambda x: datetime,
            sqlalchemy_type.Text: self.handle_string_type,
            sqlalchemy_type.Time: lambda x: time,
            sqlalchemy_type.TupleType: self.handle_tuple_type,
            sqlalchemy_type.Unicode: self.handle_string_type,
            sqlalchemy_type.UnicodeText: self.handle_string_type,
            sqlalchemy_type.VARBINARY: self.handle_string_type,
            sqlalchemy_type.VARCHAR: self.handle_string_type,
            # firebird
            firebird.CHAR: self.handle_string_type,
            firebird.VARCHAR: self.handle_string_type,
            # mssql
            mssql.BIT: lambda x: bool,
            mssql.DATETIME2: lambda x: datetime,
            mssql.DATETIMEOFFSET: lambda x: datetime,
            mssql.IMAGE: self.handle_string_type,
            mssql.MONEY: lambda x: Decimal,
            mssql.NTEXT: self.handle_string_type,
            mssql.REAL: self.handle_numeric_type,
            mssql.SMALLDATETIME: lambda x: datetime,
            mssql.SMALLMONEY: lambda x: Decimal,
            mssql.SQL_VARIANT: lambda x: str,
            mssql.TIME: lambda x: time,
            mssql.TINYINT: lambda x: int,
            mssql.UNIQUEIDENTIFIER: lambda x: str,
            mssql.VARBINARY: self.handle_string_type,
            mssql.XML: self.handle_string_type,
            # mysql
            mysql.BIGINT: lambda x: int,
            mysql.BIT: lambda x: bool,
            mysql.CHAR: self.handle_string_type,
            mysql.DATETIME: lambda x: datetime,
            mysql.DECIMAL: self.handle_numeric_type,
            mysql.DOUBLE: self.handle_numeric_type,
            mysql.ENUM: self.handle_enum,
            mysql.FLOAT: self.handle_numeric_type,
            mysql.INTEGER: lambda x: int,
            mysql.JSON: lambda x: Json,
            mysql.LONGBLOB: self.handle_string_type,
            mysql.LONGTEXT: self.handle_string_type,
            mysql.MEDIUMBLOB: self.handle_string_type,
            mysql.MEDIUMINT: lambda x: int,
            mysql.MEDIUMTEXT: self.handle_string_type,
            mysql.NCHAR: self.handle_string_type,
            mysql.NUMERIC: self.handle_numeric_type,
            mysql.NVARCHAR: self.handle_string_type,
            mysql.REAL: self.handle_numeric_type,
            mysql.SET: lambda x: set,
            mysql.SMALLINT: lambda x: int,
            mysql.TEXT: self.handle_string_type,
            mysql.TIME: lambda x: time,
            mysql.TIMESTAMP: lambda x: datetime,
            mysql.TINYBLOB: self.handle_string_type,
            mysql.TINYINT: lambda x: int,
            mysql.TINYTEXT: self.handle_string_type,
            mysql.VARCHAR: self.handle_string_type,
            mysql.YEAR: lambda x: conint(ge=1901, le=2155),
            # oracle
            oracle.BFILE: self.handle_string_type,
            oracle.BINARY_DOUBLE: self.handle_numeric_type,
            oracle.BINARY_FLOAT: self.handle_numeric_type,
            oracle.DATE: lambda x: datetime,  # supports time
            oracle.DOUBLE_PRECISION: self.handle_numeric_type,
            oracle.INTERVAL: lambda x: timedelta,
            oracle.LONG: self.handle_string_type,
            oracle.NCLOB: self.handle_string_type,
            oracle.NUMBER: self.handle_numeric_type,
            oracle.RAW: self.handle_string_type,
            oracle.VARCHAR2: self.handle_string_type,
            oracle.VARCHAR: self.handle_string_type,
            # postgresql
            postgresql.ARRAY: self.handle_list_type,
            postgresql.BIT: lambda x: bool,
            postgresql.BYTEA: self.handle_string_type,
            postgresql.CIDR: lambda x: Union[IPv4Network, IPv6Network],
            postgresql.DATERANGE: lambda x: Tuple[date, date],
            postgresql.DOUBLE_PRECISION: self.handle_numeric_type,
            postgresql.ENUM: self.handle_enum,
            postgresql.HSTORE: lambda x: Dict[str, str],
            postgresql.INET: lambda x: Union[IPv4Network, IPv6Network],
            postgresql.INT4RANGE: lambda x: Tuple[int, int],
            postgresql.INT8RANGE: lambda x: Tuple[int, int],
            postgresql.INTERVAL: lambda x: timedelta,
            postgresql.JSON: lambda x: Json,
            postgresql.JSONB: lambda x: Json,
            postgresql.MACADDR: lambda x: constr(regex=r"^([A-F0-9]{2}:){5}[A-F0-9]{2}$"),
            postgresql.MONEY: lambda x: Decimal,
            postgresql.NUMRANGE: lambda x: Tuple[Union[int, float], Union[int, float]],
            postgresql.TIME: lambda x: time,
            postgresql.TIMESTAMP: lambda x: datetime,
            postgresql.TSRANGE: lambda x: Tuple[datetime, datetime],
            postgresql.TSTZRANGE: lambda x: Tuple[datetime, datetime],
            postgresql.UUID: lambda x: UUID,
            # sqlite
            sqlite.DATE: lambda x: date,
            sqlite.DATETIME: lambda x: datetime,
            sqlite.JSON: lambda x: Json,
            sqlite.TIME: lambda x: time,
            # sybase
            sybase.BIT: lambda x: bool,
            sybase.IMAGE: self.handle_string_type,
            sybase.MONEY: lambda x: Decimal,
            sybase.SMALLMONEY: lambda x: Decimal,
            sybase.TINYINT: lambda x: int,
            sybase.UNICHAR: self.handle_string_type,
            sybase.UNITEXT: self.handle_string_type,
            sybase.UNIVARCHAR: self.handle_string_type,
        }

    def get_pydantic_type(self, column_type: Any) -> Any:
        """
        Given a Column.type value, return a type supported by pydantic
        """

        column_type_class = column_type if isclass(column_type) else column_type.__class__
        if issubclass(column_type_class, TypeEngine):
            try:
                provider = self.providers_map[column_type_class]
                return provider(column_type)
            except KeyError as e:
                raise ImproperlyConfiguredException("Unsupported Column type, please extend the provider table.") from e
        return type(column_type)

    def parse_model(self, model_class: DeclarativeMeta) -> Mapper:
        """
        Validates that the passed in model_class is an SQLAlchemy declarative model, and returns a Mapper of it
        """
        if not isinstance(model_class, DeclarativeMeta):
            raise ImproperlyConfiguredException(
                "Unsupported 'model_class' kwarg: " "only subclasses of the SQLAlchemy `DeclarativeMeta` are supported"
            )
        return inspect(model_class)

    def to_pydantic_model_class(self, model_class: DeclarativeMeta, **kwargs: Any) -> Type[BaseModel]:
        """
        Generates a pydantic model for a given SQLAlchemy declarative table and any nested relations.
        """

        mapper = self.parse_model(model_class=model_class)
        model_name = mapper.class_.__qualname__
        if model_name not in self.model_namespace_map:
            field_definitions: Dict[str, Any] = {}
            for name, column in mapper.columns.items():
                if column.default is not None:
                    field_definitions[name] = (
                        self.get_pydantic_type(column.type),
                        column.default if not callable(column.default) else column.default(),
                    )
                elif column.nullable:
                    field_definitions[name] = (self.get_pydantic_type(column.type), None)
                else:
                    field_definitions[name] = (self.get_pydantic_type(column.type), ...)
            related_entity_classes: List[DeclarativeMeta] = []
            if mapper.relationships:
                # list of refernces to other entities, not the self entity
                # to avoid duplication of pydantic models, we are using forward refs
                # see: https://pydantic-docs.helpmanual.io/usage/postponed_annotations/
                for name, relationship_property in mapper.relationships.items():
                    related_entity_class = relationship_property.mapper.class_
                    related_model_name = related_entity_class.__qualname__
                    if relationship_property.uselist:
                        field_definitions[name] = (List[related_model_name], ...)  # type: ignore
                    else:
                        field_definitions[name] = (related_model_name, ...)
                    # if the names are not identical, these are different SQLAlchemy entities
                    if related_model_name != model_name and related_model_name not in self.model_namespace_map:
                        related_entity_classes.append(related_entity_class)
            self.model_namespace_map[model_name] = create_model(model_name, **field_definitions)
            for related_entity_class in related_entity_classes:
                self.to_pydantic_model_class(model_class=related_entity_class)
        model = self.model_namespace_map[model_name]
        model.update_forward_refs(**self.model_namespace_map)
        return model

    def from_pydantic_model_instance(self, model_class: DeclarativeMeta, pydantic_model_instance: BaseModel) -> Any:
        """
        Create an instance of a given model_class using the values stored in the given pydantic_model_instance
        """
        return model_class(**pydantic_model_instance.dict())

    def to_dict(self, model_instance: Any) -> Dict[str, Any]:
        """
        Given a model instance, convert it to a dict of values that can be serialized
        """
        model_class = model_instance.__class__
        pydantic_model = self.model_namespace_map.get(model_class.__qualname__) or self.to_pydantic_model_class(
            model_class=model_class
        )
        kwargs: Dict[str, Any] = {}
        for field in pydantic_model.__fields__:
            kwargs[field] = getattr(model_instance, field)
        return pydantic_model(**kwargs).dict()



from typing import Any, Dict, List, NamedTuple, Optional, TypeVar

from pydantic import BaseModel
from typing_extensions import Protocol, Type, get_args, runtime_checkable

T = TypeVar("T")


@runtime_checkable
class PluginProtocol(Protocol[T]):  # pragma: no cover
    def to_pydantic_model_class(self, model_class: Type[T], **kwargs: Any) -> Type[BaseModel]:  # pragma: no cover
        """
        Given a model_class T, convert it to a subclass of the pydantic BaseModel
        """
        ...

    @staticmethod
    def is_plugin_supported_type(value: Any) -> bool:
        """
        Given a value of indeterminate type, determine if this value is supported by the plugin.
        """
        ...

    def from_pydantic_model_instance(self, model_class: Type[T], pydantic_model_instance: BaseModel) -> T:
        """
        Given an instance of a pydantic model created using a plugin's 'to_pydantic_model_class',
        return an instance of the class from which that pydantic model has been created.

        This class is passed in as the 'model_class' kwarg.
        """
        ...

    def to_dict(self, model_instance: T) -> Dict[str, Any]:
        """
        Given an instance of a model supported by the plugin, return a dictionary of serilizable values.
        """
        ...


def get_plugin_for_value(value: Any, plugins: List[PluginProtocol]) -> Optional[PluginProtocol]:
    """Helper function to returns a plugin to handle a given value, if any plugin supports it"""
    if value and isinstance(value, (list, tuple)):
        value = value[0]
    if get_args(value):
        value = get_args(value)[0]
    for plugin in plugins:
        if plugin.is_plugin_supported_type(value):
            return plugin
    return None


class PluginMapping(NamedTuple):
    plugin: PluginProtocol
    model_class: Any



from typing import Dict, List, Optional, Union

from openapi_schema_pydantic import (
    Contact,
    ExternalDocumentation,
    Info,
    License,
    OpenAPI,
    PathItem,
    Reference,
    SecurityRequirement,
    Server,
    Tag,
)
from pydantic import AnyUrl, BaseModel
from typing_extensions import Type

from starlite.openapi.controller import OpenAPIController


class CORSConfig(BaseModel):
    allow_origins: List[str] = ["*"]
    allow_methods: List[str] = ["*"]
    allow_headers: List[str] = ["*"]
    allow_credentials: bool = False
    allow_origin_regex: Optional[str] = None
    expose_headers: List[str] = []
    max_age: int = 600


class OpenAPIConfig(BaseModel):
    """Class containing Settings and Schema Properties"""

    create_examples: bool = False
    openapi_controller: Type[OpenAPIController] = OpenAPIController

    title: str
    version: str
    contact: Optional[Contact] = None
    description: Optional[str] = None
    external_docs: Optional[ExternalDocumentation] = None
    license: Optional[License] = None
    security: Optional[List[SecurityRequirement]] = None
    servers: List[Server] = [Server(url="/")]
    summary: Optional[str] = None
    tags: Optional[List[Tag]] = None
    terms_of_service: Optional[AnyUrl] = None
    webhooks: Optional[Dict[str, Union[PathItem, Reference]]] = None

    def to_openapi_schema(self) -> OpenAPI:
        """Generates an OpenAPI model"""
        return OpenAPI(
            externalDocs=self.external_docs,
            security=self.security,
            servers=self.servers,
            tags=self.tags,
            webhooks=self.webhooks,
            info=Info(
                title=self.title,
                version=self.version,
                description=self.description,
                contact=self.contact,
                license=self.license,
                summary=self.summary,
                termsOfService=self.terms_of_service,
            ),
        )



import re
from abc import ABC
from inspect import isclass
from typing import Any, Dict, ItemsView, List, Optional, Tuple, Union, cast

from pydantic import validate_arguments
from pydantic.typing import AnyCallable
from starlette.datastructures import URLPath
from starlette.routing import BaseRoute as StarletteBaseRoute
from starlette.routing import (
    Match,
    NoMatchFound,
    compile_path,
    get_name,
    replace_params,
)
from starlette.types import Receive, Scope, Send
from typing_extensions import Type

from starlite.controller import Controller
from starlite.enums import HttpMethod, ScopeType
from starlite.exceptions import ImproperlyConfiguredException, MethodNotAllowedException
from starlite.handlers import BaseRouteHandler, HTTPRouteHandler, WebsocketRouteHandler
from starlite.provide import Provide
from starlite.request import Request, WebSocket
from starlite.response import Response
from starlite.types import ControllerRouterHandler, Guard, Method, ResponseHeader
from starlite.utils import find_index, join_paths, normalize_path, unique

param_match_regex = re.compile(r"{(.*?)}")


class BaseRoute(ABC, StarletteBaseRoute):
    __slots__ = (
        "app",
        "handler_names",
        "methods",
        "param_convertors",
        "path",
        "path_format",
        "path_parameters",
        "path_regex",
        "scope_type",
    )

    @validate_arguments()
    def __init__(
        self,
        *,
        handler_names: List[str],
        path: str,
        scope_type: ScopeType,
        methods: Optional[List[Method]] = None,
    ):
        if not path.startswith("/"):
            raise ImproperlyConfiguredException("Routed paths must start with '/'")
        self.handler_names = handler_names
        self.path = path
        self.scope_type = scope_type
        self.path_regex, self.path_format, self.param_convertors = compile_path(path)
        self.path_parameters: List[str] = param_match_regex.findall(self.path)

        self.methods = methods or []
        if "GET" in self.methods:
            self.methods.append("HEAD")
        for parameter in self.path_parameters:
            if ":" not in parameter or not parameter.split(":")[1]:
                raise ImproperlyConfiguredException("path parameter must declare a type: '{parameter_name:type}'")

    @property
    def is_http_route(self) -> bool:
        """Determines whether the given route is an http or websocket route"""
        return self.scope_type == "http"

    def matches(self, scope: Scope) -> Tuple[Match, Scope]:
        """
        Try to match a given scope's path to self.path
        """
        if scope["type"] == self.scope_type.value:
            match = self.path_regex.match(scope["path"])
            if match:
                matched_params = match.groupdict()
                for key, value in matched_params.items():
                    matched_params[key] = self.param_convertors[key].convert(value)
                path_params = dict(scope.get("path_params", {}))
                path_params.update(matched_params)
                child_scope = {"endpoint": self, "path_params": path_params}
                if self.is_http_route and scope["method"] not in self.methods:
                    return Match.PARTIAL, child_scope
                return Match.FULL, child_scope
        return Match.NONE, {}

    def url_path_for(self, name: str, **path_params: str) -> URLPath:
        seen_params = set(path_params.keys())
        expected_params = set(self.param_convertors.keys())

        if name not in self.handler_names or seen_params != expected_params:
            raise NoMatchFound()

        path, remaining_params = replace_params(self.path_format, self.param_convertors, path_params)
        assert not remaining_params
        return URLPath(path=path, protocol=self.scope_type.value)


class HTTPRoute(BaseRoute):
    __slots__ = (
        "app",
        "handler_names",
        "methods",
        "param_convertors",
        "path",
        "path_format",
        "path_parameters",
        "path_regex",
        "route_handler_map",
        "scope_type",
    )

    @validate_arguments()
    def __init__(
        self,
        *,
        path: str,
        route_handlers: Union[HTTPRouteHandler, List[HTTPRouteHandler]],
    ):
        route_handlers = route_handlers if isinstance(route_handlers, list) else [route_handlers]
        self.route_handler_map = self.parse_route_handlers(route_handlers=route_handlers, path=path)
        super().__init__(
            methods=[method.to_str() for method in self.route_handler_map],
            path=path,
            scope_type=ScopeType.HTTP,
            handler_names=[get_name(cast(AnyCallable, route_handler.fn)) for route_handler in route_handlers],
        )

    @staticmethod
    def parse_route_handlers(route_handlers: List[HTTPRouteHandler], path: str) -> Dict[HttpMethod, HTTPRouteHandler]:
        """
        Parses the passed in route_handlers and returns a mapping of http-methods and route handlers
        """
        mapped_route_handlers: Dict[HttpMethod, HTTPRouteHandler] = {}
        for route_handler in route_handlers:
            for http_method in route_handler.http_methods:
                if mapped_route_handlers.get(http_method):
                    raise ImproperlyConfiguredException(
                        f"handler already registered for path {path!r} and http method {http_method}"
                    )
                mapped_route_handlers[http_method] = route_handler
        return mapped_route_handlers

    async def handle(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        ASGI app that creates a Request from the passed in args, and then awaits a Response
        """
        if scope["method"] not in self.methods:
            raise MethodNotAllowedException()
        request: Request[Any, Any] = Request(scope=scope, receive=receive, send=send)
        request_method = HttpMethod.from_str(request.method)
        handler = self.route_handler_map[request_method]
        response = await handler.handle_request(request=request)
        await response(scope, receive, send)


class WebSocketRoute(BaseRoute):
    __slots__ = (
        "app",
        "handler_names",
        "methods",
        "param_convertors",
        "path",
        "path_format",
        "path_parameters",
        "path_regex",
        "route_handler",
        "scope_type",
    )

    @validate_arguments()
    def __init__(
        self,
        *,
        path: str,
        route_handler: WebsocketRouteHandler,
    ):
        self.route_handler = route_handler
        super().__init__(
            path=path,
            scope_type=ScopeType.WEBSOCKET,
            handler_names=[get_name(cast(AnyCallable, route_handler.fn))],
        )

    async def handle(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        ASGI app that creates a WebSocket from the passed in args, and then awaits the handler function
        """
        web_socket: WebSocket[Any, Any] = WebSocket(scope=scope, receive=receive, send=send)
        await self.route_handler.handle_websocket(web_socket=web_socket)


class Router:
    @validate_arguments(config={"arbitrary_types_allowed": True})
    def __init__(
        self,
        *,
        path: str,
        route_handlers: List[ControllerRouterHandler],
        dependencies: Optional[Dict[str, Provide]] = None,
        guards: Optional[List[Guard]] = None,
        response_class: Optional[Type[Response]] = None,
        response_headers: Optional[Dict[str, ResponseHeader]] = None,
    ):
        self.owner: Optional["Router"] = None
        self.routes: List[BaseRoute] = []
        self.path = normalize_path(path)
        self.response_class = response_class
        self.dependencies = dependencies
        self.response_headers = response_headers
        self.guards = guards
        for route_handler in route_handlers or []:
            self.register(value=route_handler)

    @property
    def route_handler_method_map(self) -> Dict[str, Union[WebsocketRouteHandler, Dict[HttpMethod, HTTPRouteHandler]]]:
        """
        Returns dictionary that maps paths (keys) to a list of route handler functions (values)
        """
        route_map: Dict[str, Union[WebsocketRouteHandler, Dict[HttpMethod, HTTPRouteHandler]]] = {}
        for route in self.routes:
            if isinstance(route, HTTPRoute):
                if not isinstance(route_map.get(route.path), dict):
                    route_map[route.path] = {}
                for method, handler in route.route_handler_map.items():
                    route_map[route.path][method] = handler  # type: ignore
            else:
                route_map[route.path] = cast(WebSocketRoute, route).route_handler
        return route_map

    @staticmethod
    def map_route_handlers(
        value: Union[Controller, BaseRouteHandler, "Router"],
    ) -> ItemsView[str, Union[WebsocketRouteHandler, Dict[HttpMethod, HTTPRouteHandler]]]:
        """
        Maps route handlers to http methods
        """
        handlers_map: Dict[str, Union[WebsocketRouteHandler, Dict[HttpMethod, HTTPRouteHandler]]] = {}
        if isinstance(value, BaseRouteHandler):
            if isinstance(value, HTTPRouteHandler):
                handlers_map[value.path or ""] = {http_method: value for http_method in value.http_methods}
            elif isinstance(value, WebsocketRouteHandler):
                handlers_map[value.path or ""] = value
        elif isinstance(value, Router):
            handlers_map = value.route_handler_method_map
        else:
            # we are dealing with a controller
            for route_handler in value.get_route_handlers():
                path = join_paths([value.path, route_handler.path]) if route_handler.path else value.path
                if isinstance(route_handler, HTTPRouteHandler):
                    if not isinstance(handlers_map.get(path), dict):
                        handlers_map[path] = {}
                    for http_method in route_handler.http_methods:
                        handlers_map[path][http_method] = route_handler  # type: ignore
                else:
                    handlers_map[path] = cast(WebsocketRouteHandler, route_handler)
        return handlers_map.items()

    def validate_registration_value(
        self, value: ControllerRouterHandler
    ) -> Union[Controller, BaseRouteHandler, "Router"]:
        """
        Validates that the value passed to the register method is supported
        """
        if isclass(value) and issubclass(cast(Type[Controller], value), Controller):
            return cast(Type[Controller], value)(owner=self)
        if not isinstance(value, (Router, BaseRouteHandler)):
            raise ImproperlyConfiguredException(
                "Unsupported value passed to `Router.register`. "
                "If you passed in a function or method, "
                "make sure to decorate it first with one of the routing decorators"
            )
        if isinstance(value, Router):
            if value.owner:
                raise ImproperlyConfiguredException(f"Router with path {value.path} has already been registered")
            if value is self:
                raise ImproperlyConfiguredException("Cannot register a router on itself")
        else:
            # the route handler is copied to ensure each time the route handler is registerd,
            # we get an instance with a unique owner
            value = value.copy()
        value.owner = self
        return cast(Union[Controller, BaseRouteHandler, "Router"], value)

    def register(self, value: ControllerRouterHandler) -> List[Union[HTTPRouteHandler, WebsocketRouteHandler]]:
        """
        Register a Controller, Route instance or RouteHandler on the router

        Accepts a subclass or instance of Controller, an instance of Router or a function/method that has been decorated
        by any of the routing decorators (e.g. route, get, post...) exported from 'starlite.routing'
        """
        validated_value = self.validate_registration_value(value)
        handlers: List[Union[HTTPRouteHandler, WebsocketRouteHandler]] = []
        for route_path, handler_or_method_map in self.map_route_handlers(value=validated_value):
            path = join_paths([self.path, route_path])
            if isinstance(handler_or_method_map, WebsocketRouteHandler):
                handlers.append(handler_or_method_map)
                self.routes.append(WebSocketRoute(path=path, route_handler=handler_or_method_map))
            else:
                route_handlers = list(handler_or_method_map.values())
                handlers.extend(route_handlers)
                if self.route_handler_method_map.get(path):
                    existing_route_index = find_index(
                        self.routes, lambda x: x.path == path  # pylint: disable=cell-var-from-loop
                    )
                    assert existing_route_index != -1, "unable to find_index existing route index"
                    if isinstance(self.route_handler_method_map[path], dict):
                        route_handlers.extend(list(cast(dict, self.route_handler_method_map[path]).values()))
                    self.routes[existing_route_index] = HTTPRoute(
                        path=path,
                        route_handlers=unique(route_handlers),
                    )
                else:
                    self.routes.append(HTTPRoute(path=path, route_handlers=unique(route_handlers)))
        return unique(handlers)



from inspect import Signature, getfullargspec
from typing import Any, ClassVar, Dict, List, Type

from pydantic import BaseConfig, BaseModel, create_model
from pydantic.typing import AnyCallable
from pydantic_factories import ModelFactory
from typing_extensions import get_args

from starlite.exceptions import ImproperlyConfiguredException
from starlite.plugins.base import PluginMapping, PluginProtocol, get_plugin_for_value


class SignatureModel(BaseModel):
    class Config(BaseConfig):
        arbitrary_types_allowed = True

    field_plugin_mappings: ClassVar[Dict[str, PluginMapping]]
    return_annotation: ClassVar[Any]


def create_function_signature_model(fn: AnyCallable, plugins: List[PluginProtocol]) -> Type[SignatureModel]:
    """
    Creates a subclass of SignatureModel for the signature of a given function
    """

    try:
        signature = Signature.from_callable(fn)
        field_plugin_mappings: Dict[str, PluginMapping] = {}
        field_definitions: Dict[str, Any] = {}
        for key, value in getfullargspec(fn).annotations.items():
            if key == "return":
                continue
            parameter = signature.parameters[key]
            if key in ["request", "socket"]:
                # pydantic has issues with none-pydantic classes that receive generics
                field_definitions[key] = (Any, ...)
                continue
            if ModelFactory.is_constrained_field(parameter.default):
                field_definitions[key] = (parameter.default, ...)
                continue
            plugin = get_plugin_for_value(value=value, plugins=plugins)
            if plugin:
                type_args = get_args(value)
                type_value = type_args[0] if type_args else value
                field_plugin_mappings[key] = PluginMapping(plugin=plugin, model_class=type_value)
                pydantic_model = plugin.to_pydantic_model_class(model_class=type_value)
                if type_args:
                    value = List[pydantic_model]  # type: ignore
                else:
                    value = pydantic_model
            if parameter.default is not signature.empty:
                field_definitions[key] = (value, parameter.default)
            elif not repr(parameter.annotation).startswith("typing.Optional"):
                field_definitions[key] = (value, ...)
            else:
                field_definitions[key] = (value, None)
        name = (fn.__name__ if hasattr(fn, "__name__") else "anonymous") + "_signature_model"
        model: Type[SignatureModel] = create_model(name, __base__=SignatureModel, **field_definitions)
        model.return_annotation = signature.return_annotation
        model.field_plugin_mappings = field_plugin_mappings
        return model
    except TypeError as e:
        raise ImproperlyConfiguredException(repr(e)) from e
